@@ -73,11 +73,12 @@ VR.Urls = {
 };
 
 // All proc objects should subscribe to messages from this object, in the form
-// "change:<proc id>", and update themselves when such a message comes in.
+// "updateproc:<id>" and "destroyproc:<id>" and update themselves when such a
+// message comes in.
 // This saves having to drill down through collections or the DOM later if data
 // comes in on an API or event stream.
-VR.Messages = {};
-_.extend(VR.Messages, Backbone.Events);
+VR.ProcMessages = {};
+_.extend(VR.ProcMessages, Backbone.Events);
 
 
 // Base model for all models that talk to the Tastypie API. This includes
@@ -92,7 +93,10 @@ VR.Models.Tasty = Backbone.Model.extend({
 VR.Models.Proc = Backbone.Model.extend({
     initialize: function() {
       this.on('change', this.updateUrl);
-      VR.Messages.on('change:' + this.id, this.set, this);
+      VR.ProcMessages.on('updateproc:' + this.id, this.set, this);
+      VR.ProcMessages.on('destroyproc:'+this.id, this.onDestroyMsg, this); 
+
+      _.bindAll(this);
     },
 
     url: function() {
@@ -128,6 +132,10 @@ VR.Models.Proc = Backbone.Model.extend({
     getLog: function() {
       // return a ProcLog model bound to this Proc model.
       return new VR.Models.ProcLog({proc: this});
+    },
+
+    onDestroyMsg: function(data) {
+      this.trigger('destroy', this, this.collection);
     }
 });
 
@@ -188,18 +196,6 @@ VR.Models.ProcList = Backbone.Collection.extend({
       _.each(stale, function(proc) {this.remove(proc);}, this);
     },
 
-    removeByData: function(data) {
-      // Given an object like we'd get from the API or on an event stream, see if
-      // there's a proc in the collection that matches it, and remove it if so.
-
-      var proc = _.find(this.models, function(proc) {
-          return proc.id === data.id;
-        });
-      if (proc) {
-        this.remove(proc);
-      }
-    },
-
     stopAll: function() {
       this.each(function(proc) {
         proc.stop();
@@ -229,18 +225,51 @@ VR.Models.Host = VR.Models.Tasty.extend({
       var p = new VR.Models.Proc(el);
       this.procs.add(p);
     }, this);
+
+    this.procs.on('add', this.onAddProc, this);
+    this.procs.on('remove', this.onRemoveProc, this);
+  },
+
+  onProcData: function(ev, data) {
+    // backbone will instsantiate a new Proc and handle deduplication in the collection for us, 
+    // if procs have a consistent 'id' attribute, which they do.
+    this.procs.add(data);
+  },
+
+  onAddProc: function(proc) {
+    this.trigger('addproc', proc);
+  },
+
+  onRemoveProc: function(proc) {
+    this.trigger('removeproc', proc);
+    // if there are no more procs, then remove self
+    if (this.procs.length === 0 && this.collection) {
+      this.collection.remove(this);
+    }
   }
 });
 
 
 VR.Models.HostList = Backbone.Collection.extend({
-  model: VR.Models.Host
+  model: VR.Models.Host,
+  onProcData: function(ev, data) {
+    // see if we already have a host for this proc.  If not, make one.  Pass the proc down to that host.
+    var hostId = [data.app_name, data.recipe_name, data.host].join('-');
+
+    var host = this.get(hostId);
+    if (!host) {
+      host = new VR.Models.Host({ id: hostId, name: data.host });
+      this.add(host);
+    }
+    host.onProcData(ev, data);
+  }
 });
 
 
 VR.Models.Swarm = VR.Models.Tasty.extend({
     initialize: function() {
-      this.procs = new VR.Models.ProcList();
+      this.hosts = new VR.Models.HostList();
+      this.hosts.on('all', this.onHostsEvent, this);
     },
 
     procIsMine: function(fullProcName) {
@@ -272,8 +301,51 @@ VR.Models.Swarm = VR.Models.Tasty.extend({
           swarm.set(data.objects[0]);
         };
       });
-    }
+    },
 
+    onProcData: function(ev, data) {
+      // called when proc data comes down from above 
+      this.hosts.onProcData(ev, data);
+    },
+
+    onHostsEvent: function(event, model, collection, options) {
+      // all events on our hosts list should be bubbled up to be events on 
+      // the swarm itself.
+      this.trigger.apply(this, arguments);
+
+      // if all my hosts are gone, I should go too.
+      if (event === 'remove' && this.hosts.length === 0 && this.collection) {
+        this.collection.remove(this);
+      }
+    },
+
+    getProcs: function() {
+      // loop over all hosts in this.hosts and build up an array with all their procs.
+      // return that.
+      var procsArrayArray = _.map(this.hosts.models, function(host){return host.procs.models;});
+      return _.flatten(procsArrayArray);
+    },
+
+    // convenience methods since we'll so often work with procs at the swarm level.
+    // FIXME: Provide a server-side view that will restart all procs in a swarm with a
+    // single API call, instead of doing all this looping.
+    stopAll: function() {
+      this.hosts.each(function(host) {
+        host.procs.stopAll();
+      });
+    },
+
+    startAll: function() {
+      this.hosts.each(function(host) {
+        host.procs.startAll();
+      });
+    },
+
+    restartAll: function() {
+      this.hosts.each(function(host) {
+        host.procs.restartAll();
+      });
+    }
 });
 
 
@@ -295,10 +367,11 @@ VR.Models.SwarmList = Backbone.Collection.extend({
       }
       // else create, add, and return.
       swarm = new VR.Models.Swarm({
-          name: swarmname,
+          name: swarmname
         });
 
       swarm.fetchByProcData(procData);
+      swarm.on('addproc', this.onAddProc, this);
       this.add(swarm);
       return swarm;
     },
@@ -313,6 +386,13 @@ VR.Models.SwarmList = Backbone.Collection.extend({
           return swarm.procs.models.length === 0;
         }, this);
       _.each(empty_swarms, function(swarm) {this.remove(swarm);}, this);
+    },
+
+    onProcData: function(ev, data) {
+      // when we hear from above that new proc data has come in, send that down
+      // the tree.
+      var swarm = this.getByProcData(data);
+      swarm.onProcData(ev, data);
     }
 });
 
@@ -320,12 +400,33 @@ VR.Models.SwarmList = Backbone.Collection.extend({
 VR.Models.App = VR.Models.Tasty.extend({
     initialize: function() {
       this.swarms = new VR.Models.SwarmList();
+      this.swarms.on('all', this.onSwarmsEvent, this);
+    },
+
+    onProcData: function(ev, data) {
+      this.swarms.onProcData(ev, data);
+    },
+
+    onAddProc: function(proc) {
+      this.trigger('addproc', proc);
+    },
+
+    onSwarmsEvent: function(event, model, collection, options) {
+      this.trigger.apply(this, arguments);
+      if (event === 'removeproc') {
+        console.log(this.swarms);
+      }
     }
 });
 
 
 VR.Models.AppList = Backbone.Collection.extend({
     model: VR.Models.App,
+    
+    initialize: function() {
+      this.on('change', this.updateUrl);
+      VR.ProcMessages.on('all', this.onProcData, this);
+    },
 
     comparator: function(app) {
       return app.id;
@@ -351,6 +452,19 @@ VR.Models.AppList = Backbone.Collection.extend({
       return app;
     },
 
+    onProcData: function(ev, data) {
+      // handle updates differently from removals.
+      var parsed = ev.split(":");
+      if (parsed[0] === 'updateproc') {
+        var app = this.getOrCreate(data.app_name);
+        app.onProcData(ev, data);
+      } else if (parsed[0] === 'destroyproc') {
+        // TODO: here's where we should handle drilling down to remove
+        // destroyed procs, empty swarms, and empty apps, rather than from the
+        // outside in vrdash.js.
+      }
+    },
+
     cull: function(host, cutoff) {
       // call cull on each swarmlist on each swarm in the collection.
       _.each(this.models, function(app) {
@@ -409,8 +523,8 @@ VR.Views.Proc = Backbone.View.extend({
       this.modalTemplate = modalTemplate || VR.Templates.ProcModal;
 
       this.proc.on('change', this.render, this);
-      this.proc.on('destroy', this.onRemove, this);
-      this.proc.on('remove', this.onRemove, this);
+      this.proc.on('remove', this.onProcRemove, this);
+      this.proc.on('destroy', this.onProcRemove, this);
       this.render();
     },
     render: function() {
@@ -421,16 +535,16 @@ VR.Views.Proc = Backbone.View.extend({
       'click': 'onClick'
     },
 
+    onProcRemove: function(proc) {
+      this.remove();
+    },
+
     onClick: function(ev) {
       if (!this.modal) {
         this.modal = new VR.Views.ProcModal(this.proc);   
       }
 
       this.modal.show();
-    },
-
-    onRemove: function() {
-      this.$el.remove();
     }
 });
 
@@ -511,7 +625,7 @@ VR.Views.ProcModal = Backbone.View.extend({
     },
 
     onDestroyBtn: function(ev) {
-      this.proc.destroy();
+      this.proc.destroy({wait: true, sync: true});
     },
 
     onProcDestroy: function() {
@@ -556,7 +670,7 @@ VR.Views.ProcLog = Backbone.View.extend({
 
 
 VR.Views.Host = Backbone.View.extend({
-    el: '<div class="hostview"></div>',
+    el: '<tr class="hostview"></tr>',
     initialize: function(host, template) {
       this.host = host;
       this.template = template || VR.Templates.Host;
@@ -581,17 +695,42 @@ VR.Views.Host = Backbone.View.extend({
 });
 
 
+VR.Views.ProcList = Backbone.View.extend({
+    el: '<div class="procboxes"></div>',
+    initialize: function(owner) {
+      // owner is an object that may fire the 'addproc' event, which could be
+      // an app, swarm, or host.
+      this.owner = owner;
+      // whenever the owner adds a proc, add a view to ourself.
+      this.owner.on('addproc', this.addProcView, this);
+    },
+
+    addProcView: function(proc) {
+      // create a procview.
+      var pv = new VR.Views.Proc(proc);
+      // insert it in my container
+      this.$el.append(pv.el);
+    }
+});
+
+
 VR.Views.Swarm = Backbone.View.extend({
-    el: '<div class="swarmbox"></div>',
+    el: '<tr class="swarmbox"></tr>',
     initialize: function(swarm, template) {
       this.swarm = swarm;
       this.template = template || VR.Templates.Swarm;
+      // what a new host gets added, make sure we render a hostview for it
+      this.swarm.hosts.on('add', this.hostAdded, this);
+
       // when a new proc is added, make sure it gets rendered in here
-      this.swarm.procs.on('add', this.procAdded, this);
+      this.swarm.hosts.on('addproc', this.procAdded, this);
       this.swarm.on('remove', this.onRemove, this);
+
+      this.procsEl = this.$el.find('.procgrid');
     },
     events: {
-      'click .swarmtitle': 'onClick'
+      'click .swarmtitle': 'onClick',
+      'click th .expandtree': 'toggleExpanded'
     },
     onClick: function(ev) {
       if (!this.modal) {
@@ -600,10 +739,18 @@ VR.Views.Swarm = Backbone.View.extend({
 
       this.modal.show();
     },
+    toggleExpanded: function() {
+      this.$el.toggleClass('biggened');
+      this.$el.find('i').toggleClass('icon-caret-right').toggleClass('icon-caret-down');
+    },
+    hostAdded: function(host) {
+      var hv = new VR.Views.Host(host);
+      this.$el.find('.hosttable').append(hv.el);
+    },
     procAdded: function(proc) {
       var pv = new VR.Views.Proc(proc);
       pv.render();
-      this.$el.find('.procgrid').append(pv.el);
+      this.$el.children('.procgrid').append(pv.el);
     },
     render: function() {
       this.$el.html(this.template.goatee(this.swarm.toJSON()));
@@ -616,37 +763,41 @@ VR.Views.Swarm = Backbone.View.extend({
 VR.Views.SwarmModal = Backbone.View.extend({
     initialize: function(swarm, template) {
       this.swarm = swarm;
-      this.attributes.proc_length = this.swarm.procs.length;
       this.current_state = '';
       this.template = template || VR.Templates.SwarmModal;
 
-      // we subscribe to the proc's activity here too to update the list.
-      this.swarm.procs.on('add', this.procAdded, this);
-      this.swarm.procs.on('change', this.updateState, this);
-      this.swarm.on('remove', this.onRemove, this);
+      // FIXME: if a new proc is added to the swarm/host while the modal is open
+      // will we see it?  I think we need to listen for 'addproc' on the swarm to
+      // catch that.
+      this.listenTo(this.swarm, 'remove', this.remove);
+      this.listenTo(this.swarm, 'all', this.onSwarmEvent, this);
+
+      this.swarm.hosts.each(function(host) {
+          this.listenTo(host.procs, 'all', this.updateState, this);
+      }, this);
     },
     events: {
       'click .swarm-start': 'onStartBtn',
       'click .swarm-stop': 'onStopBtn',
       'click .swarm-restart': 'onRestartBtn'
     },
+
     render: function() {
       this.$el.html(this.template.goatee(this.swarm.toJSON()));
 
-      // We trigger the procModel on render to build the current procboxes at start.
-      // using for-loop here instead of _.each for scope.
-      for (var i = 0; i < this.swarm.procs.length; i++) {
-          var procModel = this.swarm.procs.models[i];
-          this.procAdded(procModel);
-      }
-      
+      var procs = this.swarm.getProcs();
+      _.each(procs, function(proc) {
+          this.procAdded(proc);
+      }, this);
     },
+
     show: function() {
       this.render();
       this.$el.modal('show');
 
       this.updateState();
     },
+
     procAdded: function(proc) {
       var pv = new VR.Views.Proc(proc);
       // unbind the model click events inside the swarm modal.
@@ -661,14 +812,15 @@ VR.Views.SwarmModal = Backbone.View.extend({
 
       // if there are any stopped/fatal procs, add the class to show the start
       // button.
-      if (this.swarm.procs.some(function(proc) {return proc.isStopped();})) {
+      var procs = this.swarm.getProcs();
+      if (_.some(procs, function(proc) {return proc.isStopped();})) {
         this.$el.find('.modal').addClass('somestopped');
       } else {
         this.$el.find('.modal').removeClass('somestopped');
       }
 
       // if there are any running procs, add the class to show the stop button
-      if (this.swarm.procs.some(function(proc) {return proc.isRunning();})) {
+      if (_.some(procs, function(proc) {return proc.isRunning();})) {
         this.$el.find('.modal').addClass('somerunning');
       } else {
         this.$el.find('.modal').removeClass('somerunning');
@@ -679,13 +831,19 @@ VR.Views.SwarmModal = Backbone.View.extend({
       this.$el.remove();
     },
     onStartBtn: function(ev) {
-      this.swarm.procs.startAll();
+      this.swarm.startAll();
     },
     onStopBtn: function(ev) {
-      this.swarm.procs.stopAll();
+      this.swarm.stopAll();
     },
     onRestartBtn: function(ev) {
-      this.swarm.procs.restartAll();
+      this.swarm.restartAll();
+    },
+    onSwarmEvent: function(event, model, collection, options) {
+      console.log(arguments);
+    },
+    onProcEvent: function(event, model, collection, options) {
+      console.log(arguments);
     }
 
 });
@@ -703,17 +861,18 @@ VR.Views.App = Backbone.View.extend({
     swarmAdded: function(swarm) {
       var sv = new VR.Views.Swarm(swarm);
       sv.render();
-      this.$el.find('.proccell').append(sv.el);
+      this.$el.find('.swarmtable').append(sv.el);
     },
 
     events: {
-      'click .expandtree': 'toggleExpanded',
+      'click .titlecell .expandtree': 'toggleExpanded',
       'click .apptitle': 'appModal'
     },
 
     toggleExpanded: function() {
       this.$el.toggleClass('biggened');
-      this.$el.find('i').toggleClass('icon-caret-right').toggleClass('icon-caret-down');
+      // only toggle this arrow, not the ones deeper inside
+      this.$el.find('.titlecell > .expandtree > i').toggleClass('icon-caret-right').toggleClass('icon-caret-down');
     },
 
     appModal: function() {
@@ -726,6 +885,8 @@ VR.Views.App = Backbone.View.extend({
 
     render: function() {
       this.$el.html(this.template.goatee(this.app.toJSON()));
+      var plv = new VR.Views.ProcList(this.app);
+      this.$el.find('td').append(plv.el);
     },
 
     onRemove: function() {
